#include "algorithms.h"
#include <math.h>
#include <iostream>
#include <algorithm>
#include "cmsketch.h"
#include "spacesaving.h"
#include <set>
#include <ctime>
#include "matplotlibcpp.h"
namespace plt = matplotlibcpp;
using namespace std;

// RNG
mt19937 rng3((int) time(0));

/*
Return median of given vector of doubles conditioned between start and end indices, inclusive.
*/
double findMedian(vector<double> a, int start, int end) {
    double* startptr = &a[start];
    double* endptr = &a[end+1];
    double median;
    if (start == end) {
        median = a[start];
    }
    else if ((start + end) % 2 == 0) {
        nth_element(startptr, &a[(start+end)/2], endptr);
        median = a[(start+end)/2];
    }
    else {
        nth_element(startptr, &a[(start+end)/2], endptr);
        double val1 = a[(start+end)/2];
        nth_element(startptr, &a[(start+end)/2 + 1], endptr);
        double val2 = a[(start+end)/2 + 1];
        median = (val1 + val2) / 2.0;
    }
    return median;
}

double getError(const vector<double>& a, int start, int end, double val) {
    double error = 0;
    for(int i = start; i <= end; i++) {
        double diff = a[i] - val;
        if (diff >= 0)
            error += diff;
        else
            error -= diff;
    }
    return error;
}

/*
Given a set of (x,y) pairs sorted by x values, find the k-piece histogram that best fits
the points. 
*/
struct histogram fitHistogramDP(const vector<long>& x, const vector<double>& y, int k, long minDomain, long maxDomain, bool useDefaultVal) {
    struct histogram hist;
    hist.k = k;
    int n = x.size(); // number of points
    if (k == 1) { // special case
        hist.leftpoints.push_back(minDomain);
        hist.rightpoints.push_back(maxDomain);
        hist.heights.push_back(findMedian(y, 0, n-1));
        return hist;
    }
    long endpoints[k][n]; // DP array of right endpoint of previous line
    double errors[k][n]; // DP array of errors
    double medians[k][n];
    endpoints[0][0] = -1;
    medians[0][0] = y[0];
    // cout << "Fitting piece... ";
    for (int l = 0; l < k; l++) { // looping through # lines
        // cout << l << " ";
        errors[l][0] = 0;
        for (int i = 1; i < n; i++) { // looping through points
            if (l == 0) {
                endpoints[l][i] = -1;
                medians[l][i] = findMedian(y, 0, i);
                errors[l][i] = getError(y, 0, i, medians[l][i]);
            }
            else { 
                long minPoint = -1;
                double minMedian = 1;
                double minError = 1;
                for (int j = 0; j < i; j++) {
                    double median;
                    double error;
                    if (j == i-1) { // avoiding catastrophic cancellation
                        median = y[i];
                        error = errors[l-1][j];
                    }
                    else {
                        median = findMedian(y, j+1, i);
                        error = errors[l-1][j] + getError(y, j+1, i, median);
                    }
                    if (error <= minError) {
                        minError = error;
                        minPoint = j;
                        minMedian = median;
                    }
                }
                assert(minPoint >= 0);
                endpoints[l][i] = minPoint;
                medians[l][i] = minMedian;
                errors[l][i] = minError;
            }
        }
    }
    // Build histogram from DP solution
    vector<long> leftpoints(k);
    vector<long> rightpoints(k);
    vector<double> heights(k);
    int cur = n-1;
    if (useDefaultVal) {
        for (int l = k-1; l >= 0; l--) {
            long nextpoint = endpoints[l][cur];
            if (l == k-1)
                rightpoints[l] = x[x.size() - 1];
            else
                rightpoints[l] = x[cur];
            if (nextpoint == -1)
                leftpoints[l] = x[0];
            else
                leftpoints[l] = x[nextpoint + 1];
            heights[l] = medians[l][cur];
            cur = nextpoint;
        }
        hist.defaultVal = findMedian(y, 0, n-1);
    }
    else {
        for (int l = k-1; l >= 0; l--) {
            long nextpoint = endpoints[l][cur];
            if (l == k-1)
                rightpoints[l] = maxDomain;
            else
                rightpoints[l] = x[cur];
            if (nextpoint == -1)
                leftpoints[l] = minDomain;
            else
                leftpoints[l] = x[nextpoint] + 1;
            heights[l] = medians[l][cur];
            cur = nextpoint;
        }
        hist.defaultVal = -1;
    }
    // fill in histogram struct
    for (int l = 0; l < k; l++) {
        hist.rightpoints.push_back(rightpoints[l]);
        hist.leftpoints.push_back(leftpoints[l]);
        hist.heights.push_back(heights[l]);
    }
    return hist;
}

struct error_breakdown getApproxError(const struct histogram& hist, const vector<long>& heavyHitters, const vector<long>& support, const map<long, double>& density, const vector<double>& heavyHeights) {
    struct error_breakdown err;
    err.errTotal = 0;
    err.heavySavings = 0;
    err.errDefault = 0;
    err.errPieces.push_back(0);
    int curpiece = 0;
    int curhh = 0;
    for (long x : support) {
        // advance histogram piece
        while (x > hist.rightpoints[curpiece] && curpiece < hist.rightpoints.size() - 1) {
            curpiece += 1;
            err.errPieces.push_back(0);
        }
        // heavy hitters (no error)
        if (curhh < heavyHitters.size() && x == heavyHitters[curhh]) { 
            if (heavyHeights.empty()) {
                err.heavySavings += density.at(x);
            } else {
                double diff = density.at(x) - heavyHeights[curhh];
                if (diff >= 0)
                    err.errTotal += diff;
                else
                    err.errTotal -= diff;
            }
            curhh++;
        }
        // histogram piece
        else if (x <= hist.rightpoints[curpiece] && x >= hist.leftpoints[curpiece]) {
            double diff = density.at(x) - hist.heights[curpiece];
            if (diff >= 0) {
                err.errTotal += diff;
                err.errPieces[curpiece] += diff;
            }
            else {
                err.errTotal -= diff;
                err.errPieces[curpiece] -= diff;
            }
        }
        // histogram default
        else {
            double diff = density.at(x) - hist.defaultVal;
            if (diff >= 0) {
                err.errTotal += diff;
                err.errDefault += diff;
            }
            else {
                err.errTotal -= diff;
                err.errDefault -= diff;
            }
        }
    }
    return err;
}

/*
Runs the one pass algorithm, saves a plot of the result and returns the error.
*/
struct error_breakdown onePassAlgo(const Stream& stream, int space, int k, string plotName, bool useDefaultVal, bool verbose) {
    map<long, double> density = stream.getDensity();
    vector<long> elems = stream.getStream();
    shuffle(elems.begin(), elems.end(), rng3);
    SpaceSaving spacesave(space/2);
    for (long id : elems) {
        spacesave.update(id, 1);
    }
    vector<long> heavyHitters;
    vector<double> heavyHeights;
    double streamLength = stream.getStreamLength(); 
    for (auto elem : spacesave.getHeavy()) {
        double mass = elem.second / streamLength;
        if (mass > (4.0/space)) {
            heavyHitters.push_back(elem.first);
            heavyHeights.push_back(mass);
        }
    }
    assert(is_sorted(heavyHitters.begin(), heavyHitters.end()));

    vector<long> samplesL0 = stream.sampleL0(space/2, heavyHitters);
    vector<double> heights;
    for (long x : samplesL0) {
        heights.push_back(density[x]);
    }
    struct histogram hist = fitHistogramDP(samplesL0, heights, k, 0, stream.getDomainSize(), useDefaultVal);
    vector<long> support = stream.getSupport();

    struct error_breakdown err = getApproxError(hist, heavyHitters, support, density, heavyHeights);
    if (verbose)
        cout << "Error: " << err.errTotal << endl;
    if (!plotName.empty()) 
        plotApprox(hist, heavyHitters, samplesL0, stream, err, plotName);
    return err;
}

/*
Runs the one pass algorithm, saves a plot of the result and returns the error.
*/
struct error_breakdown onePassAlgoSampling(const Stream& stream, int numSamples0, int numSamples1, int k, string plotName, bool useDefaultVal, bool verbose) {
    map<long, double> density = stream.getDensity();
    vector<long> samplesL1 = stream.sampleL1(numSamples1);
    vector<long> samplesL0 = stream.sampleL0(numSamples0, samplesL1);
    if (verbose) {
        cout << endl << "L1 samples: " << samplesL1.size() << endl;
        cout << "L0 samples: " << samplesL0.size() << endl;
        cout << "k: " << k << endl;
    }
    vector<double> heights;
    for (long x : samplesL0) {
        heights.push_back(density[x]);
    }
    struct histogram hist = fitHistogramDP(samplesL0, heights, k, 0, stream.getDomainSize(), useDefaultVal);
    vector<long> support = stream.getSupport();
    struct error_breakdown err = getApproxError(hist, samplesL1, support, density);
    if (verbose)
        cout << "Error: " << err.errTotal << endl;
    if (!plotName.empty()) 
        plotApprox(hist, samplesL1, samplesL0, stream, err, plotName);
    return err;
}

/*
Runs the two pass algorithm, saves a plot of the result and returns the error.
Implicitly assumes that the minimum stream element is 0.
*/
struct error_breakdown twoPassAlgo(const Stream& stream, int numSamples0, int k, string plotName, bool verbose) {
    long n = stream.getDomainSize();
    long N = stream.getStreamLength();
    map<long, double> density = stream.getDensity();
    vector<long> elems = stream.getStream();
    shuffle(elems.begin(), elems.end(), rng3);
    int h = (int)log2(n-1) + 1; // depth of hierarchy: ceil(log2(n))
    vector<SpaceSaving> hierarchy(h, numSamples0/h); //TODO: check sizes
    // compute sketches
    for (long id: elems) {
        hierarchy[0].update(id, 1);
        for (int l = 1; l < h; l++) { // levels of hierarchy
            int hierID = id / pow(2, l);
            hierarchy[l].update(hierID, 1);
        }
    }
    // set threshold
    long threshSingleton = N * h / numSamples0;
    long threshHierarchy = N * h / numSamples0;
    // cout << "Thresh: " << thresh << endl;
    vector<long> heavyHitters;
    vector<long> breakpoints = {0, n}; // starting points of new pieces
    vector<map<long, long>> incexc(h); //inclusion/exclusion tracking
    for (int l = 0; l < h; l++) {
        vector<pair<long, long>> hh = hierarchy[l].getHeavy();
        for (auto item : hh) {
            long id = item.first;
            long count = item.second;
            long offset = 0;
            if (incexc[l].find(id) != incexc[l].end())
                offset = incexc[l][id];
            long thresh = threshHierarchy;
            if (l == 0)
                thresh = threshSingleton;
            if (count - offset >= thresh) { // HHH
                if (l == 0) {
                    heavyHitters.push_back(id);
                    // cout << "HH1: " << id << " " << count << " " << offset << endl;
                } else {
                    breakpoints.push_back(min<long>((id * pow(2,l)), n));
                    breakpoints.push_back(min<long>(((id+1) * pow(2,l)), n));
                    // cout << "HHH: " << id << " " << count << " " << offset << endl;
                }
                for (int par = l+1; par < h; par++) { //subtract from parents
                    int parID = id / pow(2, par-l);
                    long delta = count - offset; // marginal extra count
                    incexc[par][parID] += delta;
                }
            }
        }
    }
    sort(heavyHitters.begin(), heavyHitters.end());
    sort(breakpoints.begin(), breakpoints.end());
    breakpoints.erase(unique(breakpoints.begin(), breakpoints.end()), breakpoints.end()); // remove duplicates
    
    if (verbose)
        cout << "nHeavy: " << heavyHitters.size() << ", nPieces: " << breakpoints.size() << endl;

    // second pass, fitting medians
    struct histogram hist;
    vector<long> allSamples;
    hist.k = breakpoints.size() - 1;
    for (int i = 0; i < hist.k; i++) {
        long left = breakpoints[i];
        long right = breakpoints[i+1]-1;
        hist.leftpoints.push_back(left);
        hist.rightpoints.push_back(right);
        vector<long> samples = stream.sampleL0FromRange(numSamples0/hist.k, left, right, heavyHitters);
        if (samples.size() == 0) { // all elements heavy
            hist.heights.push_back(0);
            continue;
        }
        allSamples.insert(allSamples.end(), samples.begin(), samples.end());
        vector<double> heights;
        for (auto x : samples) {
            heights.push_back(density[x]);
        }
        hist.heights.push_back(findMedian(heights, 0, heights.size()-1));
    }
    struct error_breakdown err = getApproxError(hist, heavyHitters, stream.getSupport(), density);
    if (!plotName.empty()) 
        plotApprox(hist, heavyHitters, allSamples, stream, err, plotName);
    return err;
}

/*
*/
double optHistogram(const Stream& stream, int k, bool supportAware, string plotName, bool verbose) {
    assert(supportAware);
    map<long, double> density = stream.getDensity();
    vector<long> points;
    vector<double> heights;
    for (auto x : density) {
        points.push_back(x.first);
        heights.push_back(x.second);
    }
    struct histogram hist = fitHistogramDP(points, heights, k, 0, stream.getDomainSize(), false);
    vector<long> support = stream.getSupport();
    vector<long> heavyHitters;
    struct error_breakdown err = getApproxError(hist, heavyHitters, support, density);
    if (verbose)
        cout << "Error: " << err.errTotal << endl;
    if (!plotName.empty()) 
        plotApprox(hist, heavyHitters, points, stream, err, plotName);
    return err.errTotal;
}

/*
Splits the domain into k fixed pieces and uses numSamples0/k L0 samples from each to approximate the median.
Calculates the error and saves a plot.
*/
double fixedIntervalAlgo(const Stream& stream, int numSamples0, int k, bool supportAware, string plotName) {
    map<long, double> density = stream.getDensity();
    struct histogram hist;
    hist.k = k;
    long n = stream.getDomainSize();
    long left;
    long right;
    vector<long> heavyHitters; //empty
    vector<long> allSamples;
    for (int i = 0; i < k; i++) {
        left = i * (n/k);
        right = (i+1) * (n/k)-1; // oboe??
        hist.leftpoints.push_back(left);
        hist.rightpoints.push_back(right);
        vector<long> samples;
        if (supportAware) { // support-aware
            samples = stream.sampleL0FromRange(numSamples0/k, left, right, heavyHitters);
            if (samples.size() == 0) {
                hist.heights.push_back(0);
                continue;
            }
        }
        else { // support-oblivous
            uniform_int_distribution<long> unif(left, right);
            set<long> uniqueSamples;
            while (uniqueSamples.size() < numSamples0/k && uniqueSamples.size() < (right - left)) {
                uniqueSamples.insert(unif(rng3));
            }
            for (long samp : uniqueSamples)
                samples.push_back(samp);
            sort(samples.begin(), samples.end());
        }
        allSamples.insert(allSamples.end(), samples.begin(), samples.end());
        vector<double> heights;
        for (auto x : samples) {
            if (density.find(x) != density.end())
                heights.push_back(density[x]);
            else
                heights.push_back(0);
        }
        hist.heights.push_back(findMedian(heights, 0, heights.size()-1));
    }
    struct error_breakdown err = getApproxError(hist, heavyHitters, stream.getSupport(), density);

    if (!plotName.empty()) 
        plotApprox(hist, heavyHitters, allSamples, stream, err, plotName);
    return err.errTotal;
}

double trueMedianAlgo(const Stream& stream, string plotName) {
    map<long, double> density = stream.getDensity();
    vector<double> heights;
    for (auto x : density) {
        heights.push_back(x.second);
    }
    double median = findMedian(heights, 0, heights.size()-1);
    struct histogram hist;
    hist.k = 1;
    hist.leftpoints.push_back(0);
    hist.rightpoints.push_back(stream.getDomainSize());
    hist.heights.push_back(median);
    vector<long> heavyHitters; // empty
    vector<long> samplesL0; // empty
    struct error_breakdown err = getApproxError(hist, heavyHitters, stream.getSupport(), density);
    // cout << endl << "Median: "  << median << endl;
    // cout << "Error: " << error << endl;
    plotApprox(hist, heavyHitters, samplesL0, stream, err, plotName);
    return err.errTotal;
}

void plotApprox(const struct histogram& hist, const vector<long> heavyHitters, const vector<long> samplesL0, const Stream& stream, struct error_breakdown err, string outname, bool logscale) {
    map<long, double> density = stream.getDensity();
    plt::figure_size(1200, 800);
    vector<double> heavyHeights;
    for (long x: heavyHitters) {
        double val = 0;
        if (density.find(x) != density.end())
            val = density.at(x);
        if (logscale)
            heavyHeights.push_back(log(val));
        else
            heavyHeights.push_back(val);
    }
    vector<double> lightHeights;
    for (long x: samplesL0) {
        double val = 0;
        if (density.find(x) != density.end())
            val = density.at(x);
        if (logscale)
            lightHeights.push_back(log(val));
        else
            lightHeights.push_back(val);
    }
    plt::scatter(heavyHitters, heavyHeights, 6.0);
    plt::scatter(samplesL0, lightHeights, 6.0, {{"c", "orange"}});
    for (int i = 0; i < hist.k; i++) {
        vector<long> xhist = {hist.leftpoints[i], hist.rightpoints[i]};
        vector<double> yhist(2);
        if (logscale) {
            yhist[0] = log(hist.heights[i]);
            yhist[1] = log(hist.heights[i]);
        }
        else {
            yhist[0] = hist.heights[i];
            yhist[1] = hist.heights[i];
        }
        plt::plot(xhist, yhist, "r--");
        plt::scatter(xhist, yhist, 6.0, {{"c", "red"}});
    }
    if (hist.defaultVal > 0) {
        vector<long> x = {0, stream.getDomainSize()};
        vector<double> y;
        if (logscale) 
            y = {log(hist.defaultVal), log(hist.defaultVal)};
        else
            y = {hist.defaultVal, hist.defaultVal};
        plt::plot(x, y, "k--");
    }
    if (logscale) {
        plt::ylim(log(stream.getMinDensity())-0.5, log(stream.getMaxDensity())+0.5);
        plt::title("log plot (error: " + to_string(err.errTotal) + ")");
    }
    else {
        plt::ylim(stream.getMinDensity(), stream.getMaxDensity());
        plt::title("error: " + to_string(err.errTotal));
    }
    plt::save(outname);
    plt::close();
}

void plotDensity(const Stream& stream, string outname, bool logscale) {
    map<long, double> density = stream.getDensity();
    vector<long> support = stream.getSupport();
    plt::figure_size(1200, 800);
    vector<double> heights;
    for (long x : support) {
        if (logscale)
            heights.push_back(log(density.at(x)));
        else
            heights.push_back(density.at(x));
    }
    plt::scatter(support, heights, 5.0);
    if (logscale)
        plt::title("log density");
    else
        plt::title("density");
    plt::save(outname);
    plt::close();
}

// USING CMSKETCH
// /*
// Runs the two pass algorithm, saves a plot of the result and returns the error.
// Implicitly assumes that the minimum stream element is 0.
// */
// struct error_breakdown twoPassAlgo(const Stream& stream, int w, int d, int numSamples0, int k, string plotName, bool verbose) {
//     long n = stream.getDomainSize();
//     map<long, double> density = stream.getDensity();
//     int h = (int)log2(n-1) + 1; // depth of hierarchy: ceil(log2(n))
//     vector<CMSketch> sketchTree(h, {w, d});
//     // compute sketches
//     for (auto x : density) {
//         sketchTree[0].update(x.first, x.second);
//         for (int l = 1; l < h; l++) { // levels of hierarchy
//             int id = x.first / pow(2, l);
//             sketchTree[l].update(id, x.second);
//         }
//     }
//     // hierarchical heavy hitters
//     double thresh = 2.0/(w*k);
//     // cout << thresh << endl;
//     vector<long> heavyHitters;
//     vector<long> breakpoints = {0, n}; // starting points of new pieces
//     for (int l = 0; l < h; l++) {
//         for (long id = 0; id < ceil(n / pow(2,l)); id++) {
//             if(sketchTree[l].estimate(id) > thresh) {
//                 if (l == 0) { //singletons
//                     heavyHitters.push_back(id);
//                 } else {
//                     breakpoints.push_back(min<long>((id * pow(2,l)), n));
//                     breakpoints.push_back(min<long>(((id+1) * pow(2,l)), n));
//                 }
//                 for (int par = l+1; par < h; par++) { //subtract from parents
//                     int parID = id / pow(2, par-l);
//                     sketchTree[par].update(parID, -sketchTree[l].estimate(id));
//                 }
//             }
//         }
//     }
//     sort(breakpoints.begin(), breakpoints.end());
//     breakpoints.erase(unique(breakpoints.begin(), breakpoints.end()), breakpoints.end()); // remove duplicates
    
//     if (verbose)
//         cout << "nHeavy: " << heavyHitters.size() << ", nPieces: " << breakpoints.size() << endl;

//     // second pass, fitting medians
//     struct histogram hist;
//     vector<long> allSamples;
//     hist.k = breakpoints.size() - 1;
//     for (int i = 0; i < hist.k; i++) {
//         long left = breakpoints[i];
//         long right = breakpoints[i+1]-1;
//         cout << left << ", " << right <<endl;
//         hist.leftpoints.push_back(left);
//         hist.rightpoints.push_back(right);
//         vector<long> samples = stream.sampleL0FromRange(numSamples0/hist.k, left, right, heavyHitters);
//         if (samples.size() == 0) { // all elements heavy
//             hist.heights.push_back(0);
//             continue;
//         }
//         allSamples.insert(allSamples.end(), samples.begin(), samples.end());
//         vector<double> heights;
//         for (auto x : samples) {
//             heights.push_back(density[x]);
//         }
//         hist.heights.push_back(findMedian(heights, 0, heights.size()-1));
//     }
//     struct error_breakdown err = getApproxError(hist, heavyHitters, stream.getSupport(), density);
//     plotApprox(hist, heavyHitters, allSamples, stream, err, plotName);
//     return err;
// }