#include "stream.h"
#include <iostream>
#include <fstream>
#include <string>
#include <sstream>
#include <set>
#include <ctime>
using namespace std;

// RNG
mt19937 rng((int) time(0));

Stream::Stream(long domainSize) {
    assert(domainSize > 0);
    n = domainSize;
}

long Stream::getStreamLength() const {
    assert(m > 0);
    return m;
}

long Stream::getDomainSize() const {
    return n;
}

long Stream::getSupportSize() const {
    assert(supp > 0);
    return supp;
}

int Stream::setStream(const vector<long>& s) {
    assert(stream.size() == 0);
    assert(density.size() == 0);
    for (size_t i = 0; i < s.size(); i++) {
        stream.push_back(s[i]);
        density[s[i]]++;
    }
    m = stream.size();
    minDensity = 1;
    maxDensity = 0;
    for (auto & x : density) {
        x.second /= m;
        if (x.second < minDensity)
            minDensity = x.second;
        if (x.second > maxDensity)
            maxDensity = x.second;
        support.push_back(x.first);
    }
    shuffle(stream.begin(), stream.end(), rng); // random order
    sort(support.begin(), support.end());
    supp = support.size();
    return 0;
}

const vector<long>& Stream::getStream() const {
    return stream;
}

const map<long, double>& Stream::getDensity() const {
    return density;
}

const vector<long>& Stream::getSupport() const {
    return support;
}

double Stream::getMinDensity() const {
    return minDensity;
}

double Stream::getMaxDensity() const {
    return maxDensity;
}

void Stream::printStats() const {
    cout << "Domain size: " << n << endl;
    cout << "Support size: " << supp << endl;
    cout << "Stream length: " << m << endl;
}

vector<long> Stream::sampleL0(long numSamples, const vector<long>& heavyHitters) const {
    vector<long> samples;
    set<long> unique_samples;
    uniform_int_distribution<long> unif(0, support.size() - 1);
    for (long i = 0; i < numSamples; i++) {
        long samp = support[unif(rng)]; 
        if (unique_samples.find(samp) == unique_samples.end() && !binary_search(heavyHitters.begin(), heavyHitters.end(), samp)) {
            // don't sample duplicate keys or heavy hitters
            unique_samples.insert(samp);
            samples.push_back(samp);
        }
    }
    sort(samples.begin(), samples.end());
    return samples;
}

vector<long> Stream::sampleL0FromRange(long numSamples, long left, long right, const vector<long>& heavyHitters) const {
    auto leftptr = lower_bound(support.begin(), support.end(), left);
    auto rightptr = upper_bound(support.begin(), support.end(), right);
    vector<long> supportInRange(leftptr, rightptr);
    vector<long> samples;
    if (supportInRange.size() == 0)
        return samples;
    set<long> unique_samples;
    uniform_int_distribution<long> unif(0, supportInRange.size() - 1);

    for (long i = 0; i < numSamples; i++) {
        long samp = supportInRange[unif(rng)]; 
        if (unique_samples.find(samp) == unique_samples.end() && !binary_search(heavyHitters.begin(), heavyHitters.end(), samp)) {
            // don't sample duplicate keys or heavy hitters
            unique_samples.insert(samp);
            samples.push_back(samp);
        }
    }
    sort(samples.begin(), samples.end());
    return samples; 
}


vector<long> Stream::sampleL1(long numSamples) const {
    vector<long> samples;
    set<long> unique_samples;
    uniform_int_distribution<int> unif(0, m-1);
    for (long i = 0; i < numSamples; i++) {
        long samp = stream[unif(rng)];
        // assert(binary_search(support.begin(), support.end(), samp));
        if (unique_samples.find(samp) == unique_samples.end()) { // don't sample duplicate keys
            unique_samples.insert(samp);
            samples.push_back(samp);
        }
    }
    sort(samples.begin(), samples.end());
    return samples;
}

Stream uniformStream(long n) {
    Stream stream = Stream(n);
    vector<long> s = vector<long>(n);
    for (long i = 0; i < n; i++) {
        s[i] = i;
    }
    stream.setStream(s);
    return stream;
}

Stream randomStream(long n, long m) {
    Stream stream = Stream(n);
    vector<long> s = vector<long>(m);
    uniform_int_distribution<int> unif(0, n-1);
    for (long i = 0; i < m; i++) {
        s[i] = unif(rng);
    }
    stream.setStream(s);
    return stream;
}

Stream mixtureGaussianStream(long n, int k, int maxMean, double stdev, bool verbose) {
    if (verbose)
        cout << "Generating gaussian mixture densities... ";
    long blocksize = n / (2*k);
    vector<long> s;
    uniform_int_distribution<int> unif(0, maxMean);
    for (int i = 0; i < k; i++) {
        double mean = unif(rng);
        if (verbose)
            cout << mean << "  ";
        normal_distribution<double> dist(mean, stdev);
        for (long j = 2*i*blocksize; j < 2*i*blocksize + blocksize; j++) {
            double nreps = dist(rng);
            if (nreps > 0) {
                for (long rep = 0; rep < (long) nreps; rep++)
                    s.push_back(j);
            }
        }
    }
    Stream stream = Stream(n);
    stream.setStream(s);
    if (verbose)
        cout << "Done!" << endl;
    return stream;
}

Stream lowerBoundStream(long n, long nElephant, long nMice, long sizeElephant) {
    vector<long> s;
    uniform_int_distribution<int> unifElephant(0, 2*n/3);
    uniform_int_distribution<int> unifMice(n/3, n);
    for (long i = 0; i < nElephant; i++) {
        long val = unifElephant(rng);
        for (long j = 0; j < sizeElephant; j++)
            s.push_back(val);
    }
    for (long i = 0; i < nMice; i++) {
        s.push_back(unifMice(rng));
    }
    Stream stream = Stream(n);
    stream.setStream(s);
    return stream;
}

vector<long> readtxtcolumn(string fname, int colnum, char delim=' ') {
    vector<long> column;
    fstream file;
    file.open(fname, ios::in);
    if (file.is_open()) {
        int ctr = 0;
        string line;
        while (getline(file, line)) {
            // if (ctr % 100000 == 0 && ctr > 0)
                // cout << ctr << " ";
            stringstream linestream(line);
            string record;
            int colctr = 0;
            while (getline(linestream, record, delim)) {// process comma separated line
                if (colctr == colnum) {
                    column.push_back(stol(record));
                }
                colctr++;
            }
            ctr++;
        }
    }
    else {
        cerr << "Problem opening file: " << fname << endl;
        exit(EXIT_FAILURE);
    }
    file.close();
    return column;
}

Stream buildDataset(vector<long>& keys, vector<long>& freqs, long n, bool verbose) {
    long minkey = *min_element(keys.begin(), keys.end());
    long maxkey = *max_element(keys.begin(), keys.end());
    long minfreq = *min_element(freqs.begin(), freqs.end());
    long maxfreq = *max_element(freqs.begin(), freqs.end());
    if (verbose) {
        cout << "Done" << endl;
        cout << "Min Key: " << minkey << ", Max Key: " << maxkey << endl;
        cout << "Min Freq: " << minfreq << ", Max Freq: " << maxfreq << endl;
        cout << "Building stream... ";
    }
    if (n <= 0)
        n = maxkey - minkey + 1; //domain size not given
    Stream stream = Stream(n);
    vector<long> s;
    for (long i = 0; i < keys.size(); i++) {
        for (long j = 0; j < freqs[i]; j++)
            s.push_back(keys[i]);
    }
    stream.setStream(s);
    if (verbose) 
        cout << "Done" << endl;
    return stream;
}

Stream importCaida(string type, int nchunks, bool verbose) {
    string fname = "../data/caida/txt/20160121-130000.txt";
    if (verbose)
        cout << "Loading caida data: " << fname << "...   " << endl;

    vector<long> keys;
    long n = -1;
    if (type == "destport") {
        keys = readtxtcolumn(fname, 9);
        n = pow(256,2);
    }
    else if (type == "sourceip" or type == "destip") {
        n = pow(256,nchunks);
        int offset = 0;
        if (type == "destip")
            offset = 4;
        for (int i = 0; i < nchunks; i++) {
            vector<long> chunk = readtxtcolumn(fname, offset + i);
            if (i == 0) {
                for (long j = 0; j < chunk.size(); j++)
                    keys.push_back(chunk[j] * (long) pow(256, nchunks-1));
            }
            else {
                assert(keys.size() == chunk.size());
                for (long j = 0; j < keys.size(); j++)
                    keys[j] += (chunk[j] * (long) pow(256, nchunks-1-i));
            }
        }
    } else {
        cout << "Unknown type " << type << endl;
        exit(0);
    }
    vector<long> freqs = readtxtcolumn(fname, 11);
    return buildDataset(keys, freqs, n, verbose);
}

Stream importMcDonalds(string type, bool verbose) {
    string fname1 = "../data/mcdonalds/"+type+"Keys.txt";
    string fname2 = "../data/mcdonalds/"+type+"Freqs.txt";
    if (verbose)
        cout << "Loading mcdonalds data: " << fname1 << "...   " << endl;

    vector<long> keys = readtxtcolumn(fname1, 0);
    vector<long> freqs = readtxtcolumn(fname2, 0);
    return buildDataset(keys, freqs, -1, verbose);
}


Stream importStanford(bool verbose) {
    string fname = "../data/stanford_ip/stanford_ip.csv";
    if (verbose)
        cout << "Loading stanford data: " << fname << "...   " << endl;

    vector<long> keys = readtxtcolumn(fname, 2, ',');
    vector<long> freqs = readtxtcolumn(fname, 3, ',');
    return buildDataset(keys, freqs, -1, verbose);
}

Stream importHamlet(bool verbose) { //trigram prefix frequency in hamlet
    string fname = "../data/hamlet/densityPrefix.txt";
    if (verbose)
        cout << "Loading hamlet data: " << fname << "...   " << endl;

    vector<long> keys(pow(26,3));
    for (int i = 0; i < pow(26,3); i++) {
        keys[i] = i;
    }
    vector<long> freqs = readtxtcolumn(fname, 0, ',');
    return buildDataset(keys, freqs, -1, verbose);
}

Stream importWP(bool verbose) { //trigram prefix frequency in hamlet
    string fname = "../data/warpeace/densityPrefix.txt";
    if (verbose)
        cout << "Loading war and peace data: " << fname << "...   " << endl;

    vector<long> keys(pow(26,3));
    for (int i = 0; i < pow(26,3); i++) {
        keys[i] = i;
    }
    vector<long> freqs = readtxtcolumn(fname, 0, ',');
    return buildDataset(keys, freqs, -1, verbose);
}

Stream importTaxi(bool verbose) { 
    string fname = "../data/taxi/pickups_2021-01_week.txt";
    if (verbose)
        cout << "Loading taxi data: " << fname << "...   " << endl;

    vector<long> stream = readtxtcolumn(fname, 0, ',');
    long maxval = 7 * 24 * 60 * 60;
    Stream s = Stream(maxval);
    s.setStream(stream);
    return s;
}